/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/


use std::path::{Path, PathBuf};

use anyhow::{bail, Context};
use clap::{crate_authors, crate_description, crate_version, Clap};

use env_logger::Env;
use yama::commands::{fully_integrate_pointer_node, load_pile_descriptor, open_pile};
use yama::operations::checking::VacuumMode;
use yama::operations::{checking, extracting};
use yama::pile::{Pile, PileDescriptor, RawPile};
use yama::{commands, debug};

#[derive(Clap)]
#[clap(version = crate_version!(), author = crate_authors!(), about = crate_description!())]
struct Opts {
    /// Chooses a different pile to be the working pile.
    /// If specified, must be the name of a remote in yama.toml.
    // TODO OBS? #[clap(short, long)]
    // with: Option<String>,

    #[clap(subcommand)]
    command: PileCommand,
}

#[derive(Clap)]
enum PileCommand {
    /// Initialise a yama pile in this directory.
    Init {},
    /// Store files into the yama pile, under a named profile.
    Store {
        // TODO NEEDS ADVANCEMENT
    },
    /// Retrieve a pointer from the yama pile, using a named pointer name.
    Retrieve {
        /// Name of the pointer to retrieve.
        pointer_name: String,

        /// Limited expression(s) of files to retrieve.
        #[clap(short, long)]
        subset: Vec<PathBuf>,

        destination: PathBuf,

        /// Number of extraction workers to use. Ideal value varies, but probably not much more than
        /// the number of CPU threads.
        #[clap(long)]
        num_workers: Option<u8>,
    },
    /// Check this yama pile for corruption.
    Check {
        #[clap(long)]
        apply_gc: bool,

        #[clap(long)]
        dry_run_gc: bool,

        #[clap(long)]
        deep: bool,

        #[clap(long)]
        shallow: bool,
    },
    /// Enter a debug prompt for manually operating on the yama pile.
    Debug,
}

fn main() -> anyhow::Result<()> {
    env_logger::Builder::from_env(Env::default().default_filter_or("info")).init();

    let opts: Opts = Opts::parse();

    let open_pile = || -> anyhow::Result<(PileDescriptor, Pile<Box<dyn RawPile>>)> {
        let this_dir = Path::new(".");
        let descriptor =
            load_pile_descriptor(this_dir).context("Failed to load pile descriptor")?;
        let pile = open_pile(this_dir, &descriptor).context("Failed to open pile")?;
        Ok((descriptor, pile))
    };

    match &opts.command {
        PileCommand::Store { .. } => {
            /*let (pdesc, pile) = open_pile()?;
            let root_dir = pdesc
                .profiles
                .get(profile_name)
                .ok_or_else(|| anyhow!("Profile not found."))?;

            let new_pointer_name = get_pointer_name_at(profile_name.clone(), Utc::now());

            info!("Will save as {:?}", new_pointer_name);

            if pile.read_pointer(&new_pointer_name)?.is_some() {
                bail!("{:?} already exists. Stopping.", new_pointer_name);
            }

            let root_node = tree::scan(root_dir)?
                .ok_or_else(|| anyhow!("{:?} doesn't seem to exist.", root_dir))?;

            let mut parent: Option<String> = None;
            let prefix = get_pointer_parent_search_prefix(profile_name.clone());
            for pointer in pile.list_pointers()?.iter() {
                if pointer.starts_with(&prefix) {
                    match parent.as_ref() {
                        None => {
                            parent = Some(pointer.to_owned());
                        }
                        Some(cur_parent) => {
                            if cur_parent < pointer {
                                parent = Some(pointer.to_owned());
                            }
                        }
                    }
                }
            }

            store_fully(&pile, &root_dir, &new_pointer_name, root_node, parent)?;
             */
            unimplemented!()
        }
        PileCommand::Retrieve {
            pointer_name,
            subset,
            destination,
            num_workers: workers,
        } => {
            let (_pdesc, pile) = open_pile()?;
            let mut pointer = pile
                .read_pointer(pointer_name)?
                .expect("No pointer by that name!"); // todo be nice

            if destination.exists() {
                bail!("The destination already exists. Overwriting not allowed (yet).");
            }

            let mut root_tree_node =
                commands::retrieve_tree_node(&pile, pointer.chunk_ref.clone())?;

            fully_integrate_pointer_node(&pile, &mut root_tree_node.node, &mut pointer)?;

            // todo allow disabling apply metadata
            extracting::extract(
                destination,
                &mut root_tree_node.node,
                &pile,
                true,
                workers.unwrap_or(2),
                true,
                true,
                true,
            )?;
        }
        PileCommand::Check {
            apply_gc,
            dry_run_gc,
            deep,
            shallow,
        } => {
            let vacuum_mode = if *dry_run_gc {
                VacuumMode::DryRunVacuum
            } else if *apply_gc {
                VacuumMode::Vacuum
            } else {
                VacuumMode::NoVacuum
            };
            let (_pdesc, pile) = open_pile()?;
            let error_count = if *deep {
                checking::check_deep(pile, vacuum_mode, true)?
            } else if *shallow {
                checking::check_shallow(pile, vacuum_mode, true, true)?
            } else {
                bail!("You need to specify either --shallow or --deep.");
            };

            if error_count > 0 {
                eprintln!("THERE ARE {} ERRORS.", error_count);
            }
        }
        PileCommand::Init {} => {
            commands::init(".".as_ref())?;
        }

        PileCommand::Debug => {
            let (pdesc, pile) = open_pile()?;
            debug::debug_prompt(pdesc, pile)?;
        }
    }

    Ok(())
}
