/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/


use std::collections::hash_map::Entry;
use std::collections::{HashMap, VecDeque};
use std::convert::{TryFrom, TryInto};
use std::fs;
use std::fs::{File, OpenOptions};
use std::io::{Read, Seek, SeekFrom, Write};
use std::path::{Path, PathBuf};
use std::sync::{Arc, Condvar, Mutex};

use anyhow::{bail, Context};
use byteorder::{BigEndian, ReadBytesExt, WriteBytesExt};
use log::warn;
use nix::unistd::sync;
use rusqlite::{params, Error};
use rusqlite::{Connection, OptionalExtension};

use crate::definitions::ChunkId;
use crate::pile::{Keyspace, RawPile};
use crate::utils::bytes_to_hexstring;
use rusqlite::ffi::ErrorCode::ConstraintViolation;

/// Bloblogs will not be reused if they are already 2 GiB large.
pub const MAX_BLOBLOG_REUSE_SIZE: u64 = 2 * 1024 * 1024 * 1024;

/// A file storing a log of blobs.
/// Format:
///     Repeated:
///         <32 byte ChunkId><u32: length><length × u8: data>
pub struct Bloblog {
    pub file: File,
}

impl Bloblog {
    pub fn open(path: &Path) -> anyhow::Result<Bloblog> {
        let file = OpenOptions::new()
            .read(true)
            .write(true)
            .create(true)
            .truncate(false)
            .open(path)
            .with_context(|| format!("Trying to open bloblog: {:?}", path))?;

        Ok(Bloblog { file })
    }

    pub fn blob_len(&mut self, offset: u64, chunk_id: &ChunkId) -> anyhow::Result<u32> {
        self.file.seek(SeekFrom::Start(offset))?;
        let mut chunk_id_verif_buf: ChunkId = Default::default();
        self.file.read_exact(&mut chunk_id_verif_buf)?;
        if &chunk_id_verif_buf != chunk_id {
            bail!(
                "ChunkId of blob did not match; read {} wanted {}",
                bytes_to_hexstring(&chunk_id_verif_buf),
                bytes_to_hexstring(chunk_id)
            );
        }
        Ok(self.file.read_u32::<BigEndian>()?)
    }

    pub fn read_blob(
        &mut self,
        offset: u64,
        chunk_id: &ChunkId,
        buf: &mut Vec<u8>,
    ) -> anyhow::Result<()> {
        buf.clear();

        self.file.seek(SeekFrom::Start(offset))?;
        let mut chunk_id_verif_buf: ChunkId = Default::default();
        self.file.read_exact(&mut chunk_id_verif_buf)?;
        if &chunk_id_verif_buf != chunk_id {
            bail!(
                "ChunkId of blob did not match; read {} wanted {}",
                bytes_to_hexstring(&chunk_id_verif_buf),
                bytes_to_hexstring(chunk_id)
            );
        }

        let blob_length = self.file.read_u32::<BigEndian>()?;

        let start = buf.len();

        for _ in 0..blob_length {
            // add enough space on the end of the buffer to read the data
            buf.push(0);
        }

        self.file.read_exact(&mut buf[start..])?;

        Ok(())
    }

    /// Writes a blob to the end of this bloblog. Returns the offset of the blob, as needed to read
    /// it with read_blob.
    pub fn write_blob(&mut self, chunk_id: &ChunkId, data: &[u8]) -> anyhow::Result<u64> {
        let position = self.file.seek(SeekFrom::End(0))?;
        self.file.write_all(chunk_id)?;
        self.file.write_u32::<BigEndian>(data.len().try_into()?)?;
        self.file.write_all(data)?;
        Ok(position)
    }

    pub fn filesize(&mut self) -> anyhow::Result<u64> {
        Ok(self.file.seek(SeekFrom::End(0))?)
    }
}

pub type BloblogId = u32;

pub struct Inner {
    next_bloblog_id: BloblogId,
    writer_bloblogs: Vec<BloblogId>,
    open_bloblogs: HashMap<BloblogId, Arc<Mutex<Bloblog>>>, // TODO want an LRU cache with a weak hashmap...?
    connection: Connection,
    writers_in_progress: u16,
}

pub struct SqliteBloblogPile {
    inner: Arc<Mutex<Inner>>,
    path: PathBuf,
    writers_reach_zero: Condvar,
}

pub struct BloblogPointer {
    bloblog: BloblogId,
    offset: u64,
}

impl SqliteBloblogPile {
    pub fn open(path: &Path) -> anyhow::Result<SqliteBloblogPile> {
        let path = &path.join("bloblog");
        let is_new = !path.is_dir();
        if is_new {
            fs::create_dir(path)?;
        }

        let connection = Connection::open(path.join("index.db"))?;

        if is_new {
            connection.execute_batch(
                "CREATE TABLE chunks (
                    chunk_id BLOB PRIMARY KEY,
                    bloblog INT NOT NULL,
                    offset INT NOT NULL
                ) WITHOUT ROWID;

                CREATE TABLE deleted (
                    bloblog INT NOT NULL,
                    offset INT NOT NULL,
                    size INT NOT NULL,
                    PRIMARY KEY (bloblog, offset)
                ) WITHOUT ROWID;

                CREATE TABLE pointers (
                    key BLOB PRIMARY KEY,
                    data BLOB NOT NULL
                );",
            )?;
        }

        Ok(SqliteBloblogPile {
            inner: Arc::new(Mutex::new(Inner {
                next_bloblog_id: 0,
                writer_bloblogs: Vec::new(),
                open_bloblogs: HashMap::new(),
                connection,
                writers_in_progress: 0,
            })),
            path: path.to_owned(),
            writers_reach_zero: Default::default(),
        })
    }

    fn open_bloblog(&self, bloblog_id: BloblogId) -> anyhow::Result<Arc<Mutex<Bloblog>>> {
        let mut inner = self.inner.lock().unwrap();
        Ok(match inner.open_bloblogs.entry(bloblog_id) {
            Entry::Occupied(entry) => entry.get().clone(),
            Entry::Vacant(entry) => {
                let bloblog = Arc::new(Mutex::new(Bloblog::open(
                    &self.path.join(&bloblog_id.to_string()),
                )?));
                entry.insert(bloblog.clone());
                bloblog
            }
        })
    }

    fn get_writing_bloblog(&self) -> anyhow::Result<(BloblogId, Arc<Mutex<Bloblog>>)> {
        let mut inner = self.inner.lock().unwrap();
        let writing_bloblog_id: BloblogId = match inner.writer_bloblogs.pop() {
            None => {
                loop {
                    let pre_inc = inner.next_bloblog_id;
                    inner.next_bloblog_id += 1;

                    // only start writing here if it doesn't already exist!
                    let bloblog_path = &self.path.join(&pre_inc.to_string());
                    if !bloblog_path.exists() {
                        break pre_inc;
                    }
                }
            }
            Some(id) => id,
        };

        let result = Ok((
            writing_bloblog_id,
            match inner.open_bloblogs.entry(writing_bloblog_id) {
                Entry::Occupied(entry) => entry.get().clone(),
                Entry::Vacant(entry) => {
                    let bloblog = Arc::new(Mutex::new(Bloblog::open(
                        &self.path.join(&writing_bloblog_id.to_string()),
                    )?));
                    entry.insert(bloblog.clone());
                    bloblog
                }
            },
        ));
        inner.writers_in_progress += 1;
        result
    }

    /// Should be called once the bloblog has been finished writing to for the moment.
    /// Will return it to the pool if there is any space left, otherwise it will close the file.
    fn return_writing_bloblog(
        &self,
        id: BloblogId,
        bloblog: Arc<Mutex<Bloblog>>,
    ) -> anyhow::Result<()> {
        let size = bloblog.lock().unwrap().filesize()?;
        let mut inner = self.inner.lock().unwrap();
        if size < MAX_BLOBLOG_REUSE_SIZE {
            inner.writer_bloblogs.push(id);
        }
        inner.writers_in_progress -= 1;
        if inner.writers_in_progress == 0 {
            self.writers_reach_zero.notify_all();
        }
        Ok(())
    }

    fn get_chunk_pointer(&self, chunk_id: &ChunkId) -> anyhow::Result<Option<BloblogPointer>> {
        let inner = self.inner.lock().unwrap();
        Ok(inner
            .connection
            .query_row(
                "SELECT bloblog, offset FROM chunks WHERE chunk_id = ?1",
                params![&chunk_id[..]],
                |row| {
                    Ok(BloblogPointer {
                        bloblog: row.get(0)?,
                        offset: row.get::<_, i64>(1)? as u64,
                    })
                },
            )
            .optional()?)
    }

    fn put_chunk_pointer(&self, chunk_id: &ChunkId, pointer: BloblogPointer) -> anyhow::Result<()> {
        let inner = self.inner.lock().unwrap();
        let offset_i64 = i64::try_from(pointer.offset).expect("ouch! can't turn u64 into i64...");
        match inner.connection.execute(
            "INSERT INTO chunks (chunk_id, bloblog, offset) VALUES (?1, ?2, ?3)",
            params![&chunk_id[..], pointer.bloblog, offset_i64],
        ) {
            Ok(_) => Ok(()),
            Err(Error::SqliteFailure(e, str)) => {
                if e.code == ConstraintViolation {
                    warn!(
                        "(ignoring) SQLite constraint violation on insertion... {:?}",
                        str
                    );
                    Ok(())
                } else {
                    Err(Error::SqliteFailure(e, str))?;
                    unreachable!();
                }
            }
            other => {
                other?;
                unreachable!();
            }
        }
    }
}

impl RawPile for SqliteBloblogPile {
    fn exists(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<bool> {
        match kind {
            Keyspace::Chunk => {
                let mut chunk_id: ChunkId = Default::default();
                chunk_id.copy_from_slice(key);
                Ok(self.get_chunk_pointer(&chunk_id)?.is_some())
            }
            Keyspace::ChunkHash => unimplemented!(),
            Keyspace::Pointer => {
                let inner = self.inner.lock().unwrap();
                let pointer_exists = inner
                    .connection
                    .query_row(
                        "SELECT data FROM pointers WHERE key=?1",
                        params![key],
                        |_row| Ok(()),
                    )
                    .optional()?
                    .is_some();
                Ok(pointer_exists)
            }
        }
    }
    fn read(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<Option<Vec<u8>>> {
        match kind {
            Keyspace::Chunk => {
                let mut chunk_id: ChunkId = Default::default();
                chunk_id.copy_from_slice(key);
                let chunk_pointer = self.get_chunk_pointer(&chunk_id)?;
                if let Some(pointer) = chunk_pointer {
                    let bloblog_mutex = self.open_bloblog(pointer.bloblog)?;
                    let mut bloblog = bloblog_mutex.lock().unwrap();

                    let mut buf = Vec::new();
                    bloblog.read_blob(pointer.offset, &chunk_id, &mut buf)?;
                    Ok(Some(buf))
                } else {
                    Ok(None)
                }
            }
            Keyspace::ChunkHash => unimplemented!(),
            Keyspace::Pointer => {
                let inner = self.inner.lock().unwrap();
                let pointer_data = inner
                    .connection
                    .query_row(
                        "SELECT data FROM pointers WHERE key=?1",
                        params![key],
                        |row| row.get::<_, Vec<u8>>(0),
                    )
                    .optional()?;
                Ok(pointer_data)
            }
        }
    }
    fn write(&self, kind: Keyspace, key: &[u8], value: &[u8]) -> anyhow::Result<()> {
        match kind {
            Keyspace::Chunk => {
                // need to handle deletion of existing, same-named chunks
                self.delete(kind, key)?;

                let mut chunk_id: ChunkId = Default::default();
                chunk_id.copy_from_slice(key);

                let (bloblog_id, bloblog) = self.get_writing_bloblog()?;
                let offset = bloblog.lock().unwrap().write_blob(&chunk_id, value)?;
                self.return_writing_bloblog(bloblog_id, bloblog)?;
                let pointer = BloblogPointer {
                    bloblog: bloblog_id,
                    offset,
                };

                self.put_chunk_pointer(&chunk_id, pointer)?;
                Ok(())
            }
            Keyspace::ChunkHash => unimplemented!(),
            Keyspace::Pointer => {
                let inner = self.inner.lock().unwrap();
                inner.connection.execute(
                    "INSERT OR REPLACE INTO pointers (key, data) VALUES (?1, ?2)",
                    params![key, value],
                )?;
                Ok(())
            }
        }
    }
    fn delete(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<()> {
        match kind {
            Keyspace::Chunk => {
                let mut chunk_id: ChunkId = Default::default();
                chunk_id.copy_from_slice(key);
                let chunk_pointer = self.get_chunk_pointer(&chunk_id)?;
                if let Some(pointer) = chunk_pointer {
                    let bloblog_mutex = self.open_bloblog(pointer.bloblog)?;
                    let mut bloblog = bloblog_mutex.lock().unwrap();

                    let size = bloblog.blob_len(pointer.offset, &chunk_id)?;

                    let inner = self.inner.lock().unwrap();

                    let offset_i64 =
                        i64::try_from(pointer.offset).expect("ouch! can't turn u64 into i64...");

                    inner.connection.execute(
                        "INSERT OR IGNORE INTO deleted (bloblog, offset, size)
                            VALUES (?1, ?2, ?3)",
                        params![pointer.bloblog, offset_i64, size],
                    )?;
                }
                Ok(())
            }
            Keyspace::ChunkHash => unimplemented!(),
            Keyspace::Pointer => {
                let inner = self.inner.lock().unwrap();
                inner
                    .connection
                    .execute("DELETE FROM pointers WHERE key=?1", params![key])?;
                Ok(())
            }
        }
    }
    fn list_keys(
        &self,
        kind: Keyspace,
    ) -> anyhow::Result<Box<dyn Iterator<Item = anyhow::Result<Vec<u8>>>>> {
        let query = match kind {
            Keyspace::Chunk => "SELECT chunk_id FROM chunks WHERE chunk_id > ?1",
            Keyspace::ChunkHash => unimplemented!(),
            Keyspace::Pointer => "SELECT key FROM pointers WHERE key > ?1",
        };

        let ki = KeyIterator {
            next_limit: Vec::with_capacity(0),
            query,
            buffer: Default::default(),
            connection_inner: self.inner.clone(),
        };

        Ok(Box::new(ki))
    }
    fn flush(&self) -> anyhow::Result<()> {
        let inner = self.inner.lock().unwrap();
        let _inner = self
            .writers_reach_zero
            .wait_while(inner, |inner| inner.writers_in_progress != 0)
            .unwrap();
        // TODO sync all files we have open first!
        sync();
        Ok(())
    }
    fn check_lowlevel(&self) -> anyhow::Result<bool> {
        unimplemented!()
    }
}

struct KeyIterator {
    next_limit: Vec<u8>,
    query: &'static str,
    buffer: VecDeque<Vec<u8>>,
    connection_inner: Arc<Mutex<Inner>>,
}

impl KeyIterator {
    fn next_result(&mut self) -> anyhow::Result<Option<Vec<u8>>> {
        match self.buffer.pop_front() {
            None => {
                let conn = &self.connection_inner.lock().unwrap().connection;
                let mut stmt = conn.prepare(self.query)?;
                let new_rows = stmt
                    .query_map(params![&self.next_limit], |row| row.get::<_, Vec<u8>>(0))?
                    .map(|row| row.unwrap());
                self.buffer.extend(new_rows);

                if let Some(next_limit) = self.buffer.back() {
                    self.next_limit = next_limit.clone();
                }

                Ok(self.buffer.pop_front())
            }
            Some(ele) => Ok(Some(ele)),
        }
    }
}

impl Iterator for KeyIterator {
    type Item = anyhow::Result<Vec<u8>>;

    fn next(&mut self) -> Option<Self::Item> {
        match self.next_result() {
            Ok(Some(vec)) => Some(Ok(vec)),
            Ok(None) => None,
            Err(err) => Some(Err(err)),
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::pile::local_sqlitebloblogs::Bloblog;
    use temp_dir::TempDir;

    #[test]
    pub fn bloblog_read_write_test() {
        let td = TempDir::new().unwrap();
        let path = td.path();

        let mut bloblog = Bloblog::open(path.join("bloblog1.log").as_ref()).unwrap();
        let bloblog_id_1 = bloblog.write_blob(&Default::default(), &[1, 2, 3]).unwrap();
        let bloblog_id_2 = bloblog.write_blob(&Default::default(), &[4, 5]).unwrap();
        drop(bloblog);

        let mut bloblog = Bloblog::open(path.join("bloblog1.log").as_ref()).unwrap();
        let mut vec = Vec::new();
        bloblog
            .read_blob(bloblog_id_2, &Default::default(), &mut vec)
            .unwrap();
        assert_eq!(&vec, &[4, 5]);

        bloblog
            .read_blob(bloblog_id_1, &Default::default(), &mut vec)
            .unwrap();
        assert_eq!(&vec, &[1, 2, 3]);
    }
}
