use std::collections::HashMap;
use std::io::{Read, Write};
use std::sync::{Arc, Mutex};
use std::thread;
use std::thread::JoinHandle;

use anyhow::anyhow;
use crossbeam_channel::{Receiver, Sender};
use log::error;

use crate::pile::{Keyspace, RawPile};
use crate::remote::{read_message, write_message, Request, RequestBody, Response, ResponseBody};

pub struct Requester {
    commands: Sender<(RequestBody, Sender<ResponseBody>)>,
}

impl Requester {
    pub fn new<R: Read + Send + 'static, W: Write + Send + 'static>(
        read: R,
        write: W,
    ) -> (Self, Vec<JoinHandle<()>>) {
        let in_flight: Arc<Mutex<HashMap<u16, Sender<ResponseBody>>>> =
            Arc::new(Mutex::new(HashMap::new()));
        let (command_sender, command_receiver) = crossbeam_channel::bounded(16);
        let mut handles = Vec::new();

        {
            // Spawn a reader
            let in_flight = in_flight.clone();
            handles.push(thread::spawn(move || {
                if let Err(e) = Self::reader(read, in_flight) {
                    error!("reader failed: {:?}", e);
                }
            }));
        }

        {
            // Spawn a writer
            let in_flight = in_flight.clone();
            let command_receiver = command_receiver.clone();
            handles.push(thread::spawn(move || {
                if let Err(e) = Self::writer(write, in_flight, command_receiver) {
                    error!("writer failed: {:?}", e);
                }
            }));
        }

        (
            Requester {
                commands: command_sender,
            },
            handles,
        )
    }

    /// Thread that reads messages and sends them along.
    fn reader<R: Read + Send + 'static>(
        mut read: R,
        in_flight: Arc<Mutex<HashMap<u16, Sender<ResponseBody>>>>,
    ) -> anyhow::Result<()> {
        loop {
            let response: Response = read_message(&mut read)?;

            let map = in_flight.lock().or(Err(anyhow!("Mutex poisoned")))?;
            map.get(&response.response_to)
                .ok_or(anyhow!("Didn't find response channel..."))?
                .send(response.body)
                .or(Err(anyhow!("Failed to send response to channel")))?;
        }
    }

    /// Thread that writes messages.
    fn writer<W: Write + Send + 'static>(
        mut write: W,
        in_flight: Arc<Mutex<HashMap<u16, Sender<ResponseBody>>>>,
        command_receiver: Receiver<(RequestBody, Sender<ResponseBody>)>,
    ) -> anyhow::Result<()> {
        while let Ok((req_body, response_channel)) = command_receiver.recv() {
            let request_id = {
                let mut map = in_flight.lock().or(Err(anyhow!("Mutex poisoned")))?;
                let request_id = (0u16..u16::max_value())
                    .into_iter()
                    .find(|id| !map.contains_key(&id))
                    .expect("No ID found");
                map.insert(request_id, response_channel);
                request_id
            };
            write_message(
                &mut write,
                &Request {
                    id: request_id,
                    body: req_body,
                },
            )?;
        }
        Ok(())
    }

    /// Helper to make a request and wait for the result.
    fn request(&self, req: RequestBody) -> anyhow::Result<ResponseBody> {
        let (sender, receiver) = crossbeam_channel::bounded(0);
        self.commands
            .send((req, sender))
            .or(Err(anyhow!("Failed to queue request")))?;

        Ok(receiver
            .recv()
            .or(Err(anyhow!("Failed to receive response")))?)
    }
}

impl RawPile for Requester {
    fn exists(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<bool> {
        match self.request(RequestBody::CheckExists {
            kind,
            key: key.to_vec(),
        })? {
            ResponseBody::Success => Ok(true),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Ok(false),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for exists.")),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for exists.")),
        }
    }
    fn read(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<Option<Vec<u8>>> {
        match self.request(RequestBody::Read {
            kind,
            key: key.to_vec(),
        })? {
            ResponseBody::Success => Err(anyhow!("Received Success for read.")),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Ok(None),
            ResponseBody::Data(data) => Ok(Some(data)),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for read.")),
        }
    }
    fn write(&self, kind: Keyspace, key: &[u8], value: &[u8]) -> anyhow::Result<()> {
        match self.request(RequestBody::Write {
            kind,
            key: key.to_vec(),
            value: value.to_vec(),
        })? {
            ResponseBody::Success => Ok(()),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Err(anyhow!("Received NotExists for write.")),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for write.")),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for write.")),
        }
    }
    fn delete(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<()> {
        match self.request(RequestBody::Delete {
            kind,
            key: key.to_vec(),
        })? {
            ResponseBody::Success => Ok(()),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Err(anyhow!("Received NotExists for delete.")),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for delete.")),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for delete.")),
        }
    }
    fn list_keys(
        &self,
        kind: Keyspace,
    ) -> anyhow::Result<Box<dyn Iterator<Item = anyhow::Result<Vec<u8>>>>> {
        match self.request(RequestBody::List { kind })? {
            ResponseBody::Success => Ok(Box::new(ListKeyIterator {
                command_sender: self.commands.clone(),
                batch_token: None,
                buffer: Vec::with_capacity(0),
            })),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Err(anyhow!("Received NotExists for list_keys.")),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for list_keys.")),
            ResponseBody::BatchData { batch, next_token } => Ok(Box::new(ListKeyIterator {
                command_sender: self.commands.clone(),
                batch_token: Some(next_token),
                buffer: batch,
            })),
        }
    }
    fn flush(&self) -> anyhow::Result<()> {
        match self.request(RequestBody::Flush)? {
            ResponseBody::Success => Ok(()),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Err(anyhow!("Received NotExists for Flush.")),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for Flush.")),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for Flush.")),
        }
    }
    fn check_lowlevel(&self) -> anyhow::Result<bool> {
        match self.request(RequestBody::LowLevelCheck)? {
            ResponseBody::Success => Ok(true),
            ResponseBody::Failed(err_msg) => Err(anyhow!("Remote failure: {}", err_msg)),
            ResponseBody::NotExists => Err(anyhow!("Received NotExists for LowLevelCheck.")),
            ResponseBody::Data(_) => Err(anyhow!("Received Data for LowLevelCheck.")),
            ResponseBody::BatchData { .. } => Err(anyhow!("Received BatchData for LowLevelCheck.")),
        }
    }
}

pub struct ListKeyIterator {
    pub(crate) command_sender: Sender<(RequestBody, Sender<ResponseBody>)>,
    pub(crate) batch_token: Option<u16>,
    /// in reverse order
    pub(crate) buffer: Vec<Vec<u8>>,
}

impl Iterator for ListKeyIterator {
    type Item = anyhow::Result<Vec<u8>>;

    fn next(&mut self) -> Option<Self::Item> {
        if let Some(next) = self.buffer.pop() {
            Some(Ok(next))
        } else if let Some(batch_token) = self.batch_token {
            let (send, recv) = crossbeam_channel::bounded(0);
            self.command_sender
                .send((RequestBody::NextBatch { token: batch_token }, send))
                .expect("Unable to send");
            let resp = recv.recv().expect("Unable to recv");
            match resp {
                ResponseBody::Success => {
                    self.batch_token = None;
                    None
                }
                ResponseBody::Failed(err_msg) => Some(Err(anyhow!("Remote failure: {}", err_msg))),
                ResponseBody::NotExists => Some(Err(anyhow!("Received NotExists for NextBatch."))),
                ResponseBody::Data(_) => Some(Err(anyhow!("Received Data for NextBatch."))),
                ResponseBody::BatchData { batch, next_token } => {
                    self.batch_token = Some(next_token);
                    self.buffer = batch;
                    self.buffer.reverse();

                    if let Some(next) = self.buffer.pop() {
                        Some(Ok(next))
                    } else {
                        None
                    }
                }
            }
        } else {
            None
        }
    }
}
